#!/usr/bin/env python3
# automome.py
# automome - the OTTish AUTOMEME clone
# Copyright (C) 2014  Penguin Development
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import copy
import getopt
import itertools
import os
import os.path
import random
import re
import sys

def skipuntilvowel(s):
    """Function that returns the given string from the first vowel onwards.
    Returns the whole string if it has no vowels."""
    s = str(s)
    mat = re.compile(r'[AEIOU]').search(s)
    return s[mat.start():] if mat else s

# Lambda that turns a verb of form [x]ING into [x]ER:
erify_s = lambda x: (x[:-4] + 'IER') if x.endswith('YING') else (x[:-3] + 'ER')
erify = lambda x: erify_s(str(x))

# Lambda that turns `THE [x]' into just `[x]':
skipthe_s = lambda x: x[4:] if x.startswith('THE ') else x
skipthe = lambda x: skipthe_s(str(x))

# Lambda that skips the first word:
skipword = lambda x: ' '.join(str(x).split()[1:])

# Lambda that strips the indefinite article (A/AN) from singular nouns and
# adjectives:
skipart = lambda x: skipword(x) if  x.originclass == 'adjective' \
        or (x.originclass.startswith('noun') and x.wordform == 0) \
        else x.stringform

# Lambda that strips the indefinite article from mass nouns and unique nouns
# only:
skipart_mu = lambda x: skipword(x) if x.originclass in \
        ('noun-mass', 'noun-unique') else x.stringform

# Lambda that prepends 'THE ' to words from the noun-unique class:
uniquethe = lambda x: ('THE ' + skipword(x.stringform)) if \
        x.originclass == 'noun-unique' else skipart_mu(x)

class WordChoice(object):
    """Class specifying a chosen word. Objects of this type should not be
    modified by any user-defined functions."""
    wordforms = None
    wordform = 0
    originclass = None
    stringform = None

    def __str__(self):
        """Method to generate the canonical string representation of this word
        choice. This method will return `stringform' unless it is None, in
        which case wordforms[wordform] is returned instead. If `wordforms' is
        itself None, the empty string is returned, although this not normally
        happen."""
        if self.stringform is not None:
            return self.stringform
        elif self.wordforms is not None:
            return self.wordforms[self.wordform]
        else:
            return ''

class Replaceable(object):
    """Replaceable class. Objects of this class generate words to fill the
    meme templates."""
    scandir = '.'
    classes = {}

    def __init__(self, cls, defaults=None, selector=None, postproc=None):
        """Constructor.

        `defaults' is a list of tuples of default words and relative
        probabilities; e.g. [('FOO', 0.5), ('BAR', 0.33)] means there is first
        a 50% chance of 'FOO' being used, failing that, a 33% chance of 'BAR'
        being used, and failing that, a 100% chance of normal selection being
        applied.

        `cls' is a list of classes from which a replacement word may be
        chosen; an entry for a multiword class may also be given as a tuple of
        the class name and an index; e.g. ['noun-mass', ('noun', 1)]
        would pick from mass-nouns and plural nouns (since inex 1 for
        multiform nouns is the plural form).

        `selector' is a predicate function that can be used to pre-filter
        acceptable words. Predicate functions should take one argument: a
        WordChoice object specifying the word slated for selection. The
        `stringform' attribute of the WordChoice object will be the default
        form, i.e. wordforms[wordform]. The predicate should return True if
        this choice is acceptable, or False if it isn't.  Words for which the
        predicate returns False are removed from the list of possible choices.
        If the predicate does not return True for any of the choices, the
        empty string is chosen. If `selector' is None, all words from the
        classes listed in `cls' are deemed acceptable.

        `postproc' is a postprocessing function taking one argument: a
        WordChoice object specifying the word slated for selection. The
        `stringform' attribute of the WordChoice object will be set to the
        default form of the word. The postprocessing function should return
        the desired new final form of the word. The postprocessing function
        must not modify the WordChoice object. If `postproc' is None, the word
        will be used as-is."""
        self.defaults = defaults
        self.selector = selector
        self.postproc = postproc
        self.scanclasses()
        if cls is None:
            self.cls = None
            return

        self.acceptlist = None

        self.cls = []
        for c in cls:
            if isinstance(c, str):
                if c not in self.classes:
                    raise KeyError(
                            'Mustard: word class `{}\' not found!'.format(c))
                self.cls.append((c, 0))
            else:
                if c[0] not in self.classes:
                    raise KeyError(
                            'Mustard: word class `{}\' not found!'.format(
                                c[0]))
                self.cls.append(tuple(c))

    @classmethod
    def scanclasses(cls, force = False):
        """Scans the directory Replaceable.scandir for files named *.dict.txt.
        The files' base names (i.e. the part before .dict.txt) are used as
        names for word classes; the files contain one word per line, or
        multiple word forms separated by the pipe symbol | if the file name
        ends in .dict.txt.  Whitespace is stripped off the start and
        end of each word form, and empty lines (after stripping) and lines
        starting with # are ignored.

        This method aborts immediately if it has been called previously; set
        Replaceable.classes to the empty dictionary before calling this method
        to force a rescan."""

        if len(cls.classes) != 0:
            return

        for fn in os.listdir(cls.scandir):
            ffn = os.path.join(cls.scandir, fn)
            if fn.endswith('.dict.txt') and os.path.isfile(ffn):
                f = open(ffn, 'r', encoding='UTF-8')
                # the encoding in the line above is to prevent CGI mustard
                data = [x.strip() for x in f.read().splitlines()]
                f.close()
                data = [d for d in data if d and not d.startswith('#')]
                if len(data):
                    cls.classes[fn[:-9]] = [[y.strip() for y in
                        x.split('|')] for x in data]

        cls.scanned = True

    def pick(self):
        """Word-picking method. Returns a WordChoice object specifying the
        chosen word. This method will build and cache a list of words that are
        acceptable according to self.selector."""
        c = WordChoice()
        c.originclass = 'x-defaults'
        if self.defaults is not None:
            for x in self.defaults:
                if random.random() > (1. - float(x[1])):
                    c.stringform = x[0]
                    return c

        if self.cls is None or len(self.cls) == 0:
            c.stringform = '' if self.defaults is None else \
                    random.choice(self.defaults)[0]
            return c

        if self.acceptlist is None:
            self.acceptlist = []
            for cls in self.cls:
                self.acceptlist += [(cls[0], i, j) for i in cls[1:] for j in
                        range(len(self.classes[cls[0]]))]

        while len(self.acceptlist) > 0:
            idx = random.randrange(len(self.acceptlist))
            c.originclass = self.acceptlist[idx][0]
            c.wordform = self.acceptlist[idx][1]
            c.wordforms = self.classes[c.originclass][self.acceptlist[idx][2]]
            c.stringform = c.wordforms[c.wordform]
            if self.selector is None or self.selector(c) is True:
                break
            del self.acceptlist[idx]

        if len(self.acceptlist) > 0:
            if self.postproc is not None:
                c.stringform = self.postproc(c)
        else:
            c.wordforms = None
            c.wordform = 0
            c.originclass = None
            c.stringform = None

        return c

class IndexMap(object):
    """Class that applies a function to the replacement, given the argument
    index of that replacement. E.g. let R = ('foo', IndexMap(0, f)) be the
    replacement tuple, then this will get expanded to ('foo', f('foo'))."""
    def __init__(self, idx, f = lambda x: x):
        self.index = idx
        self.func = f

    def apply(self, data):
        return self.func(data)

class Meme(object):
    """Meme class."""
    def __init__(self, template, *args):
        self.template = template
        self.args = args

    def __str__(self):
        """Returns a fully-filled version of the template."""
        try:
            args = [str(x) if not isinstance(x, (Replaceable, IndexMap)) else
                    (x.pick() if isinstance(x, Replaceable) else x) for x in
                    self.args]

            for i in range(len(args)): # Replace IndexMaps
                if isinstance(args[i], IndexMap):
                    args[i] = args[i].apply(args[args[i].index]) # ARGCEPTION

            return self.template.format(*[str(x) for x in args])
        except:
            print(">>> In template " + self.template, file=sys.stderr)
            raise


# Praparatory block to set the scan dir and seed the PRNG if automome.py is
# called directly.
if __name__ == '__main__':
    p = os.getenv('AUTOMOME_PATH')
    if (p):
        Replaceable.scandir = p
    elif not Replaceable.scandir:
        Replaceable.scandir = os.path.abspath(os.path.dirname(sys.argv[0]))

    random.seed()

# Multiform noun index specification:
# 0: singular w/ article
# 1: plural
n_singular, n_plural = range(2)

# Verb index specification:
# 0: infinitive
# 1: continuous
# 2: present
# 3: past
# 4: perfect
v_inf, v_cont, v_present, v_past, v_perfect = range(5)

# Adjective index specification:
# Positive w/article
# Comparative w/article
# Superlative
a_pos, a_comp, a_super = range(3)

# Stock replaceables. The new syntax makes selecting the old classes more
# verbose; these stock replaceables should make the most common replaceables
# easy to select. Besides, this saves memory. :)
Rs = {
    'noun commodity':
        Replaceable(['noun'], postproc=skipart),
    'noun singular':
        Replaceable(['noun', 'noun-unique'], postproc=skipart),
    'noun plural':
        Replaceable([('noun', n_plural)]),
#    'noun mass':
#        Replaceable(['noun-mass']),
    'noun singular plural':
        Replaceable([('noun', n_singular, n_plural), 'noun-unique'],
            postproc=skipart),
    'noun singular plural mass':
        Replaceable([('noun', n_singular, n_plural), 'noun-unique',
            'noun-mass'], postproc=skipart),
    'noun singular mass':
        Replaceable(['noun', 'noun-unique', 'noun-mass'], postproc=skipart),
    'noun singular mass character':
        Replaceable(['noun', 'noun-unique', 'noun-mass', 'character'],
            postproc=skipart),
    'noun plural mass':
        Replaceable([('noun', n_plural), 'noun-mass'], postproc=skipart_mu),
    'noun plural mass character':
        Replaceable([('noun', n_plural), 'noun-mass', 'character'],
            postproc=skipart_mu),
    'noun singular plural mass character':
        Replaceable([('noun', n_singular, n_plural), 'noun-unique',
            'noun-mass', 'character'], postproc=skipart),
    'noun singular plural mass adjective':
        Replaceable([('noun', n_singular, n_plural), 'noun-unique',
            'noun-mass', 'adjective'], postproc=skipart),
    'noun singular character':
        Replaceable(['noun', 'noun-unique', 'character'], postproc=skipart),
    'noun plural character':
        Replaceable([('noun', n_plural), 'character']),
    'noun commodity':
        Replaceable(['noun'], postproc=skipart),
    'noun singular article':
        Replaceable(['noun']),
    'noun singular article character':
        Replaceable(['noun', 'character']),
    'noun singular article plural mass':
        Replaceable([('noun', n_singular, n_plural), 'noun-mass'],
            postproc=skipart_mu),
    'noun singular article mass character':
        Replaceable(['noun', 'noun-mass', 'character'], postproc=skipart_mu),

    'verb intr inf':
        Replaceable(['verb-intr']),
    'verb intr cont':
        Replaceable([('verb-intr', v_cont)]),
    'verb intr present':
        Replaceable([('verb-intr', v_present)]),
    'verb intr past':
        Replaceable([('verb-intr', v_past)]),
#    'verb intr perfect':
#        Replaceable([('verb-intr', v_perfect)]),
    'verb tr inf':
        Replaceable(['verb-tr']),
    'verb tr cont':
        Replaceable([('verb-tr', v_cont)]),
    'verb tr present':
        Replaceable([('verb-tr', v_present)]),
    'verb tr past':
        Replaceable([('verb-tr', v_past)]),
#    'verb tr perfect':
#        Replaceable([('verb-tr', v_perfect)]),
    'verb inf':
        Replaceable(['verb-intr', 'verb-tr']),
    'verb cont':
        Replaceable([('verb-intr', v_cont), ('verb-tr', v_cont)]),
    'verb present':
        Replaceable([('verb-intr', v_present), ('verb-tr', v_present)]),
    'verb past':
        Replaceable([('verb-intr', v_past), ('verb-tr', v_past)]),
#    'verb perfect':
#        Replaceable([('verb-intr', v_perfect), ('verb-tr', v_perfect)]),

    'verb inf 1word':
        Replaceable(['verb-intr', 'verb-tr'],
            selector=lambda x: len(x.stringform.split()) == 1),

    'adjective':
        Replaceable(['adjective'], postproc=skipart),
    'adjective article':
        Replaceable(['adjective']),
    'adjective superlative':
        Replaceable([('adjective', a_super)]),

    'character':
        Replaceable(['character']),
    'place':
        Replaceable(['place'])
}

# Easy replacement constructor that uses varargs for the class list and kwargs
# for other other arguments:
R = lambda *x, **y: Replaceable(x, **y)
# Easy defaults-only replacement constructor:
C = lambda *x, **y: Replaceable(cls=None, defaults=x, **y)

# For the origins of these memes, see mrob27's list at
#   http://mrob.com/time/automome/ott-snowclones.txt
memes_automeme_molpish = [
    Meme('ALL YOUR {} ARE BELONG TO {}',
        Rs['noun singular'],
        Rs['noun singular plural mass character']),
    Meme('ATTACK ITS WEAK POINT FOR {} {}',
        Rs['adjective'],
        Rs['noun singular plural mass']),
    Meme('BECAUSE YOU ARE _{} {}_',
        R('adjective', defaults=[('HEADCRAB', 0.33)], postproc=skipart),
        Rs['noun singular mass']),
    Meme('BECOME AN {}',
        R(('noun', n_singular, n_plural), 'noun-mass', selector=lambda x: \
                x.originclass != 'noun' or x.wordform == n_plural \
                or x.stringform[:2] == 'A ', postproc=skipart)),
    Meme('BETTER {} MY OWN {}',
        Rs['verb tr inf'],
        Rs['noun singular mass']),
    Meme('BLESS YOU, {}. BL{}.',
        Rs['noun plural mass character'],
        IndexMap(0, skipuntilvowel)),
    Meme('CEILING {} IS WATCHING YOU {}',
        Rs['noun singular character'],
        Rs['verb inf']),
    Meme('CH*RPING {}, HOW DO THEY WORK?',
        Rs['noun plural']),
    Meme('DAMNIT JIM, I\'M {}, NOT {}',
        Rs['noun singular article'],
        Rs['noun singular article']),
    Meme('DISREGARD THAT, I {} {}',
        Rs['verb tr inf'],
        Rs['noun plural mass']),
    Meme('DR. {}, OR HOW I LEARNED TO STOP {} AND LOVE THE {}',
        Rs['noun singular mass character'],
        R(('verb-intr', v_cont), defaults=[('WORRYING', 0.1)]),
        Rs['noun singular']),
    Meme('GREAT LORD RANDALL IT\'S {} GET IN THE {}',
        R('noun', 'noun-unique', 'noun-mass', 'character',
            postproc=lambda x: 'THE ' + skipword(x.stringform) \
                if x.originclass == 'noun-unique' else skipart_mu(x)),
        R('noun', 'noun-unique', 'noun-mass',
            defaults=[('CAR', 0.4), ('RAFTCASTLE', 0.4)], postproc=skipart)),
    Meme('HELLO. MY NAME IS {}. YOU {} MY {}. PREPARE TO {}.',
        Rs['character'],
        Rs['verb tr past'],
        Rs['noun singular plural mass'],
        Rs['verb intr inf']),
    Meme('HELLO? YES, THIS IS {}',
        Rs['noun singular mass character']),
    Meme('HOW DO I {} {}?',
        Rs['verb tr past'],
        Rs['noun singular plural mass character']),
    Meme('HOW IS {} FORMED? HOW {} GET {}?',
        Rs['noun singular mass'],
        Rs['noun singular mass character'],
        R('adjective', defaults=[('PRAGNENT', 0.33)], postproc=skipart)),
    Meme('I ACCIDENTALLY {}, IS THIS {}?',
        Rs['noun singular article'],
        C(('BAD', 0.5), ('DANGEROUS', 0.5), ('ENDISH', 1.0))),
    Meme('I AM _{0} {1}_, THE {0} {1} EVER!',
        R(('adjective', a_super),
            defaults=[('BEST', 0.4), ('MOLPIEST', 0.67)], selector=lambda x: \
                x.stringform.endswith('EST') \
                and not x.stringform.startswith('MOST')),
        Rs['noun singular mass']),
    Meme('I AM {}, HEAR ME {}',
        Rs['character'],
        R('verb-intr', defaults=[('ROAR', 0.5)])),
    Meme('I DON\'T ALWAYS {}, BUT WHEN I DO, I {} {}',
        Rs['verb intr inf'],
        R('verb-tr', defaults=[('PREFER', 0.2)]),
        Rs['noun plural mass']),
    Meme('I DON\'T ALWAYS {} {}, BUT WHEN I DO, I {} {}',
        Rs['verb tr inf'],
        Rs['noun plural mass'],
        R('verb-tr', defaults=[('PREFER', 0.2)]),
        Rs['noun plural mass']),
    Meme('I, FOR ONE, WELCOME OUR NEW {} OVERLORDS',
        Rs['noun singular']),
    Meme('I HAS {}. YOU CAN HAS MY {}.',
        Rs['noun singular article'],
        IndexMap(0, skipword)),
    Meme('I HAVE REPORTED YOU TO {} FOR {} {} AS IT IS A CRIME',
        Rs['character'],
        R(('verb-tr', v_cont), defaults=[('STEALING', 0.75)]),
        Rs['noun plural mass']),
    Meme('I LIKE MY {} LIKE I LIKE MY {}: COVERED IN {}',
        R(('noun', n_plural), 'noun-mass',
            defaults=[('WOMEN', 0.25), ('MEN', 0.33)], postproc=skipart_mu),
        Rs['noun plural mass'],
        Rs['noun plural mass']),
    Meme('I MADE YOU {} BUT I {} IT',
        Rs['noun singular article'],
        Rs['verb tr past']),
    Meme('IMMA CHARGIN MAH {}',
        Rs['noun singular plural mass']),
    Meme('I\'M IN UR {}, {} UR {}',
        Rs['noun singular mass'],
        Rs['verb tr cont'],
        Rs['noun plural mass']),
    Meme('IN B4 {} {}',
        R('adjective', defaults=[('503', 0.2), ('408', 0.25)],
            postproc=skipart),
        Rs['noun singular plural mass']),
    Meme('IN {}, {} {} _YOU!_',
        R('place', defaults=[('SOVIET RUSSIA', 0.5)],
            postproc=lambda x: 'THE SOVIET ' + x.stringform[4:] \
                if x.stringform.startswith('THE ') else 'SOVIET ' + \
                x.stringform),
        Rs['noun singular mass character'],
        Rs['verb tr present']),
    Meme('I SEE WHAT {} {} THERE',
        R('character', defaults=[('YOU', 0.5)]),
        Rs['verb tr past']),
    Meme('IS IT CAN BE {} TIEM NAO PLEES?',
        R(('noun', n_singular, n_plural), 'noun-unique',
            selector=lambda x: x.wordform == n_singular \
                or x.stringform.endswith('S'),
            postproc=lambda x: (x.stringform[:-1] + 'Z') \
                if x.wordform == n_plural else skipword(x))),
    Meme('I THINK {} IS A PRETTY {} {}. '
            'XEH {} {} AND DOESN\'T AFRAID OF ANYTHING',
        Rs['character'],
        R('adjective', defaults=[('TREEISH', 0.1), ('MOLPISH', 0.11)],
            postproc=skipart),
        R('noun', defaults=[('XUY', 0.1)], postproc=skipart),
        Rs['verb tr present'],
        Rs['noun plural mass character']),
    Meme('I WILL NOT {} THIS {} -- IT IS {}',
        Rs['verb tr inf'],
        Rs['noun singular mass'],
        R('noun', 'noun-unique', 'noun-mass', 'adjective', 'character',
            postproc=lambda x: skipword(x) if x.originclass == \
                'adjective' else (('THE ' + skipword(x)) if
                    x.originclass == 'noun-unique' else skipart_mu(x)))),
    Meme('I {} _ALL_ THE {}',
        Rs['verb tr past'],
        Rs['noun plural mass']),
    Meme('I {}, THEREFORE I {}',
        R('verb-intr', defaults=[('THINK', 0.5)]),
        Rs['verb intr inf']),
    Meme('MAXIMUM {} YIELDS MAXIMUM {}',
        Rs['noun plural mass'],
        Rs['noun plural mass']),
    Meme('MY {} IS FULL OF {}',
        Rs['noun singular mass'],
        Rs['noun plural mass']),
    Meme('MY {} JUST {} ME; M*STARD WAS SO {}',
        Rs['noun singular mass'],
        Rs['verb tr past'],
        R('adjective', defaults=[('CASH', 0.5)], postproc=skipart)),
    Meme('MY {}, LET ME SHOW YOU IT',
        Rs['noun singular mass']),
    Meme('NOOOO THEY BE {} MY {}',
        Rs['verb tr cont'],
        Rs['noun singular plural mass']),
    Meme('NOT SURE IF {} OR JUST {}',
        Rs['verb intr cont'],
        R(('verb-intr', v_cont), 'adjective', postproc=skipart)),
    Meme('NO, {}, YOU ARE THE {}',
        Rs['noun singular mass character'],
        Rs['noun singular mass']),
    Meme('OH GLR HOW DID {} GET HERE I AM NOT GOOD WITH {}',
        R(('noun', n_plural), 'noun-mass', 'character',
            defaults=[('THIS', 0.5)], postproc=skipart_mu),
        Rs['noun plural mass']),
    Meme('OH HAI, I {} YOUR {}',
        Rs['verb tr past'],
        Rs['noun singular plural mass']),
    Meme('PROTIP: TO {} THE {}, {} IT UNTIL IT {}',
        R('verb-tr', defaults=[('DEFEAT', 0.25)]),
        Rs['noun singular mass'],
        Rs['verb tr inf'],
        Rs['verb intr present']),
    Meme('PUSH {}, RECEIVE {}',
        C(('BUTAN', 0.5), ('BUTTON', 1.0)),
        Rs['noun singular plural mass']),
    Meme('SCREAMING {}: {} DOESN\'T LIKE IT',
        Rs['noun singular character'],
        Rs['character']),
    Meme('SO I HERD U {} {}',
        R('verb-tr', defaults=[('LIEK', 0.75)]),
        Rs['noun plural mass']),
    Meme('SO LONG AND THANKS FOR ALL THE {}',
        Rs['noun plural mass']),
    Meme('THANKS, {}. TH{}.',
        Rs['noun plural mass character'],
        IndexMap(0, skipuntilvowel)),
    Meme('THANK YOU {}, BUT OUR {} IS IN ANOTHER {}',
        Rs['character'],
        Rs['noun singular mass'],
        Rs['noun singular']),
    Meme('THAT\'S NOT {0}; _THIS_ IS {0}',
        Rs['noun singular article']),
    Meme('THERE IS {} IN MY {}, YOUR ARGUMENT IS INVALID',
        Rs['noun singular article'],
        R('noun', 'noun-unique', 'noun-mass', defaults=[('BEARD', 0.33)],
            postproc=skipart)),
    Meme('THEY SEE ME {}, THEY {}',
        R(('verb-intr', v_cont),
            selector=lambda x: x.stringform.endswith('ING'),
            postproc=lambda x: x.stringform[:-1] + '\''),
        R(('verb-intr', v_cont), defaults=[('HATIN\'', 0.33)],
            selector=lambda x: x.stringform.endswith('ING'),
            postproc=lambda x: x.stringform[:-1] + '\'')),
    Meme('THE {} IS A LIE',
        Rs['noun singular']),
    Meme('"THIS IS NOT ACCEPTABLE!" I SCREAMED AS {} {}',
        Rs['character'],
        Rs['verb intr past']),
    Meme('THIS IS YOUR {} ON {}',
        R('noun', 'noun-unique', defaults=[('BRAIN', 0.5)], postproc=skipart),
        Rs['noun plural mass']),
    Meme('THIS IS {}. I CAN TELL BY THE {}, '
            'AND FROM HAVING SEEN A LOT OF {} IN MY DIP',
        R(('verb-tr', v_perfect), 'adjective',
            defaults=[('PHOTOSHOPPED', 0.12), ('OTTIFIED', 0.14),
                ('SHOOPED', 0.12)], postproc=skipart),
        Rs['noun plural mass'],
        Rs['noun plural mass']),
    Meme('THIS IS {} {}, I MUST {} IT',
        Rs['adjective'],
        Rs['noun singular'],
        Rs['verb tr inf']),
    Meme('THIS IS {} {}. IT IS MADE OF {} AND {}.',
        Rs['adjective article'],
        Rs['noun commodity'],
        Rs['noun singular plural mass adjective'],
        Rs['noun singular plural mass adjective']),
    Meme('THIS {} IS NOW ABOUT {} {}',
        R('noun', 'noun-unique', defaults=[('NEEDLE-PULLED THING', 0.5)],
            postproc=skipart),
        Rs['adjective'],
        Rs['noun plural mass character']),
    Meme('VERY IMPORTANT FOR {} TO {} XIM {}',
        Rs['character'],
        R('verb-tr', defaults=[('DO', 0.25)]),
        R('noun', 'noun-unique', 'noun-mass', defaults=[('JOB', 0.5)],
            postproc=skipart)),
    Meme('WE\'VE SECRETLY REPLACED {}\'S {} WITH {}. '
            'LET\'S SEE IF XE NOTICES...',
        Rs['character'],
        Rs['noun singular mass'],
        Rs['noun singular article']),
    Meme('WHAT _ARE_ {}? WE JUST DON\'T KNOW.',
        Rs['noun plural']),
    Meme('WHATEVER {} YOUR {}',
        Rs['verb tr present'],
        R(('noun', n_singular, n_plural), 'noun-unique', 'noun-mass',
            defaults=[('RAFTCASTLE', 0.67), ('SHIPCASTLE', 0.33)],
            postproc=skipart)),
    Meme('WHAT THE CH*RP? YOU ARE {}. AND {}',
        Rs['noun singular mass character'],
        Rs['adjective']),
    Meme('YOU HAVE {} MOST OF A RIVERISH {}. '
            'PLEASE PICK YOUR {} WITH GREATER CARE.',
        Rs['verb tr past'],
        Rs['noun commodity'],
        Rs['noun plural mass']),
    Meme('YOU WOULDN\'T {} {}',
        R('verb-tr', defaults=[('DOWNLOAD', 0.5)]),
        Rs['noun singular article']),
    Meme('YO WAWLPY, I HEARD YOU LIKE {0} SO WE PUT {1} IN YOUR {2} '
            'SO YOU CAN {3} WHILE YOU {3}',
        IndexMap(1, lambda x: x.wordforms[n_plural] \
            if x.originclass == 'noun' else x.wordforms[x.wordform]),
        R(('noun', n_singular, n_plural), 'noun-mass', postproc=lambda x: \
            ('SOME ' + x.stringform) if x.originclass == 'noun-mass' \
            or x.wordform == n_plural else x.stringform),
        IndexMap(1, lambda x: x.wordforms[x.wordform] \
            if x.originclass == 'noun-mass' or x.wordform == n_plural \
            else skipword(x.wordforms[x.wordform])),
        Rs['verb intr inf']),
    Meme('YO {}, I\'M REALLY HAPPY FOR YOU, IMMA LET YOU FINISH, '
            'BUT {} HAD ONE OF THE {} {} OF ALL TIME',
        R('character', postproc=skipthe),
        Rs['character'],
        C(('MOST TREEISH', 0.33), ('MOLPIEST', 0.5), ('MOST WOWTERFUL', 1.0)),
        Rs['noun plural']),
    Meme('{} _ALL_ THE {}',
        Rs['verb tr inf'],
        Rs['noun plural mass']),
    Meme('{} DON\'T KNOW ABOUT MY {}',
        Rs['noun plural'],
        Rs['noun singular plural mass']),
    Meme('{}? IN _MY_ {}?',
        Rs['noun plural mass'],
        Rs['noun singular mass']),
    Meme('{} IS {} TOO',
        Rs['noun singular article'],
        Rs['adjective']),
    Meme('{} IS SERIOUS {}',
        Rs['noun singular character'],
        Rs['noun singular']),
    Meme('{} IS THE {} OF THE {}',
        Rs['noun singular character'],
        Rs['noun singular mass'],
        Rs['noun plural mass']),
    Meme('{} LOEV {}',
        R('character', defaults=[('LOLI', 0.5)]),
        Rs['noun singular plural mass character']),
    Meme('{0}MEOWLPY IS {1}',
        IndexMap(1, skipword),
        Rs['noun singular article']),
    Meme('{}\'S CLOSED DUE TO {}',
        R('noun', 'noun-unique', 'noun-mass', 'place',
            defaults=[('POOL', 0.5)], postproc=lambda x: skipthe(x) if \
                x.originclass == 'place' else skipart(x)),
        Rs['noun plural mass character']),
    Meme('{} OR GTCO',
        Rs['noun plural mass']),
    Meme('{} THE {}!',
        Rs['verb tr inf'],
        Rs['noun plural mass']),
    Meme('{} THE {}... WITH YOUR {}',
        Rs['verb tr inf'],
        Rs['noun plural mass'],
        Rs['noun singular plural mass']),
    Meme('{} {} ARE _SUPERIOR_',
        Rs['adjective'],
        Rs['noun plural'])
]

memes_other_molpish = [
    Meme('ALSO, I CAN {} YOU WITH MY {}.',
        Rs['verb tr inf'],
        Rs['noun singular']),
    Meme('AND WHY DO YOU WORRY ABOUT {}? CONSIDER THE {} OF {}, '
            'HOW THEY {}: THEY NEITHER {} NOR {}.',
        Rs['noun plural'],
        Rs['noun plural'],
        Rs['place'],
        Rs['verb tr inf'],
        Rs['verb tr inf'],
        Rs['verb tr inf']), # Credit: mrob27
    Meme('A WILD {} {}',
        Rs['noun singular'],
        C(('APPEARS', 0.5), ('APPEARED', 1.0))), # Credit: mrob27
    Meme('BRACE {}, {} IS COMING',
        C(('YOURSELF', 0.5), ('YOURSELVES', 1.0)),
        Rs['noun singular mass']), # Credit: mrob27
    Meme('CONSIDER THE {} OF {}',
        Rs['noun plural mass'],
        Rs['place']), # Credit: mrob27
    Meme('COULD {} _BE_ ANY MORE {}?',
        Rs['noun plural mass character'],
        Rs['adjective']),
    Meme('CROUCHING {}, HIDDEN {}',
        Rs['noun singular'],
        Rs['noun singular']),
    Meme('DIAL {} FOR {}',
        IndexMap(1, lambda x: x.stringform[0]),
        Rs['noun singular plural mass character']),
    Meme('DON\'T {} ME BECAUSE I\'M {}',
        R('verb-tr', defaults=[('HATE', 0.3)]),
        R('adjective', defaults=[('FLUTTERBEEWINGISH', 0.2),
            ('WINGISH', 0.25)], postproc=skipart)), # Credit: mrob27
    Meme('EVERY TIME YOU {}, {} {} {}. PLEASE, THINK OF THE {}.',
        Rs['verb intr inf'],
        R('character', defaults=[('RANDALL', 0.3), ('THE GLR', 0.43)]),
        Rs['verb tr present'],
        Rs['noun singular article'],
        IndexMap(3, lambda x: x.wordforms[n_plural])), # Credit: mrob27
    Meme('EXCUSE ME {}, DO YOU HAVE A MOMENT TO TALK ABOUT {}?',
        C(('SIR', 0.5), ('MA\'AM', 1.0)),
        R(('noun', n_plural), 'character', 'noun-unique', 'noun-mass',
            postproc=uniquethe)),
    Meme('{0} DON\'T LET {0} {1} {2}',
        R(('noun', n_plural), defaults=[('FRIENDS', 0.85)]),
        Rs['verb tr inf'],
        Rs['noun plural mass']),
    Meme('GOT {}?',
        R(('noun', n_plural), 'noun-mass', 'noun-mass', postproc=skipart)),
        # Credit: mrob27
    Meme('HAST {} THE {} {}?',
        R(('verb-tr', v_perfect), defaults=[('SEEN', 0.67)]),
        Rs['adjective'],
        R('noun', defaults=[('WHALPY', 0.33)], postproc=skipart)),
    Meme('HELLO, I\'M {}. BASICALLY, {}!',
        Rs['character'],
        R('verb-intr', defaults=[('RUN', 0.67)])),
    Meme('I CH*RPING LOVE {}',
        Rs['noun plural mass']),
    Meme('I DIDN\'T CHOOSE THE {0} LIFE; THE {0} LIFE CHOSE ME',
        Rs['noun singular mass']),
    Meme('I FIND YOUR LACK OF {} DISTURBING',
        Rs['noun plural mass']),
    Meme('IF {} {} WRONG, I DON\'T WANT TO BE RIGHT',
        R(('verb-intr', v_cont), ('verb-tr', v_cont),
            ('verb-intr', v_cont), ('verb-tr', v_cont),
            ('noun', n_plural), 'noun-unique', 'noun-mass', 'character',
            postproc=uniquethe),
        IndexMap(0, lambda x: 'ARE' if x.originclass == 'noun' else 'IS')),
        # Credit: mrob27
    Meme('IF {} {} IS {}, {} {}.',
        Rs['verb tr cont'],
        R(('noun', n_plural), 'place', defaults=[('IN YOUR PANTS', 0.3)],
            postproc=lambda x: ('IN ' if x.originclass == 'place' \
                else 'YOUR ') + x.stringform),
        Rs['adjective'],
        C(('THEN I\'M', 0.5), ('CONSIDER ME', 1.0)),
        R(('verb-tr', v_perfect), 'adjective', 'character',
            postproc=skipart)), # Credit: mrob27
    Meme('I\'M GOING TO {0}{1} LIKE I\'VE NEVER {2}{1} BEFORE.',
        Rs['verb inf'],
        IndexMap(0, lambda x: '' if x.originclass == 'verb-intr' \
            else (' ' + Rs['noun plural'].pick().stringform)),
        IndexMap(0, lambda x: x.wordforms[v_perfect])),
    Meme('I\'M SORRY {}, I\'M AFRAID I CAN\'T {} THAT.',
        Rs['character'],
        R('verb-tr', defaults=[('DO', 0.15)])),
    Meme('INSTANT {}, JUST ADD {}',
        Rs['noun singular plural mass'],
        Rs['noun plural mass']),
    Meme('IT\'S {} ALL THE WAY DOWN',
        Rs['noun plural mass']), # Credit: mrob27
    Meme('IT\'S {} GONE WILD{}!',
        Rs['noun plural'],
        C(('', 0.3), (' II', 0.3), (' III', 0.3), (' 4', 0.3), (' 5', 0.3),
            (' 6', 0.4), (' 7', 1.0))), # Credit: mrob27
    Meme('IT\'S {}. IT\'S {}. IT\'S _SUPER{}_!',
        Rs['noun singular article'],
        Rs['noun singular article'],
        R('noun', 'noun-unique', 'noun-mass',
            selector=lambda x: len(x.stringform.split()) == (2 if \
                x.originclass == 'noun' else 1), postproc=skipart)),
    Meme('I USED TO BE {} LIKE YOU, THEN I TOOK {} IN THE KNEE',
        Rs['noun singular article'],
        Rs['noun singular article']),
    Meme('I\'VE GOT A JAR OF {0}! I\'VE GOT A JAR OF {0}! '
            'AND GUESS WHAT\'S INSIDE IT!',
        Rs['noun plural mass']),
    Meme('"I WANT THE {0}!"/"YOU CAN\'T HANDLE THE {0}!"',
        Rs['noun singular mass']),
    Meme('I WEAR {} NOW. {} ARE {}.',
        Rs['noun singular article'],
        IndexMap(0, lambda x: x.wordforms[n_plural]),
        C(('COOL', 0.5), ('TREEISH', 0.5), ('MOLPISH', 1.0))),
    Meme('KEEP CALM AND {} ON',
        R('verb-intr', defaults=[('MOLPY', 0.1)],
            selector=lambda x: len(x.stringform.split()) == 1)),
        # Credit: mrob27
    Meme('LOOK ON MY {}, YE {}, AND {}!',
        Rs['noun plural mass'],
        R(('verb-intr', v_cont), 'adjective',
            selector=lambda x: x.originclass == 'adjective' or \
                (x.stringform.endswith('ING') and \
                    len(x.stringform.split()) == 1), postproc=skipart),
        R('verb-intr', defaults=[('DESPAIR', 0.25)])),
    Meme('MY {} BRINGS ALL THE {} TO {}',
        Rs['noun singular'],
        Rs['noun plural'],
        R('place', defaults=[('THE YARD', 0.5)])),
    Meme('MY NAME IS {}, {} OF {}: LOOK ON MY {}, YE {}, AND {}!',
        Rs[ 'character'],
        Rs['noun commodity'],
        IndexMap(1, lambda x: x.wordforms[n_plural]),
        Rs['noun plural mass'],
        R(('verb-intr', v_cont), 'adjective',
            selector=lambda x: x.originclass == 'adjective' or \
                (x.stringform.endswith('ING') and \
                    len(x.stringform.split()) == 1), postproc=skipart),
                R('verb-intr', defaults=[('DESPAIR', 0.25)])),
            # Credit for this form: mrob27
    Meme('NEVER SEND TO KNOW FOR WHOM THE {0} {1}; IT {1} FOR THEE',
        Rs['noun singular'],
        Rs['verb present']), # Credit: mrob27
    Meme('NOW I AM BECOME {}, THE {} OF {}',
        Rs['noun singular mass character'],
        R(('verb-tr', v_cont), selector=lambda x: \
                x.stringform.endswith('ING'), postproc=erify),
        Rs['noun plural mass']),
    Meme('NOW IF YOU\'LL EXCUSE ME, I HAVE {} TO {}',
        Rs['noun singular article plural mass'],
        Rs['verb tr inf']),
    Meme('NOW IS THE TIME FOR ALL {} {} TO COME TO THE AID OF {}',
        C(('TREEISH', 0.33), ('MOLPISH', 0.5), ('GOOD', 1.0)),
        Rs['noun plural'],
        R('noun', 'noun-unique', 'place', postproc=lambda x: \
            ('THE ' + skipart(x)) if x.originclass.startswith('noun-') \
            else x.stringform)), # Credit: mrob27
    Meme('ONE DOES NOT SIMPLY {} INTO {}',
        Rs['verb intr inf'],
        R('place', defaults=[('MORDOR', 0.3)])), # Credit: mrob27
    Meme('ONE {} DOES NOT {} MAKE',
        Rs['noun commodity'],
        Rs['noun singular article']),
    Meme('ONE {0} TO {1} THEM ALL, ONE {0} TO FIND THEM.'
            ' ONE {0} TO {2} THEM ALL AND IN {3} BIND THEM.',
        Rs['noun singular'],
        R('verb-tr', defaults=[('RULE', 0.75)]),
        R('verb-tr', defaults=[('BRING', 0.75)]),
        R('place', selector=lambda x: x.wordforms[1] == 'IN')),
    Meme('ONLY TWO THINGS COME FROM {}: {} AND {}',
        Rs['place'],
        Rs['noun plural mass character'],
        Rs['noun plural mass character']),
    Meme('PH\'NGLUI MGLW\'NAFH {} {} WGAH\'NAGL FHTAGN',
        R('character', postproc=skipthe),
        R('place', postproc=skipthe)),
    Meme('PLEASE READ: A PERSONAL APPEAL FROM {} FOUNDER {}.',
        Rs['noun singular'],
        Rs['character']), # Credit: mrob27
    Meme('PUT {} ON {}',
        Rs['noun singular mass'],
        R('noun', 'noun-unique', 'noun-mass', defaults=[('HEAD', 0.5)],
            postproc=skipart)),
    Meme('REMAIN {} AND {} {}',
        R('adjective', defaults=[('CALM', 0.2)], postproc=skipart),
        Rs['verb tr inf'],
        Rs['noun plural mass']), # Credit: mrob27
    Meme('SOMETIMES YOU {0} THE {1}, SOMETIMES THE {1} {2} YOU',
        Rs['verb tr inf'],
        Rs['noun singular'],
        IndexMap(0, lambda x: x.wordforms[v_present])),
    Meme('SPEAK {} AND CARRY {} {}',
        R('adverb', defaults=[('SOFTLY', 0.75)]),
        R('adjective', defaults=[('A SEAISH', 0.33)]),
        Rs['noun singular']),
    Meme('SUDDENLY, {}. THOUSANDS OF THEM!',
        Rs['noun plural']),
    Meme('THAT\'S ONE RIVERISH {0} FOR {1}, ONE SEAISH {0} FOR {2}KIND',
        Rs['noun singular'],
        Rs['noun singular article'],
        IndexMap(1, skipword)),
    Meme('THE FIRST RULE OF {0} IS, YOU DON\'T TALK ABOUT {0}',
        R('noun', 'noun-unique', 'noun-mass', postproc=lambda x: \
            random.choice(('THE {}', '{} CLUB')).format(skipword(x)) \
                if x.originclass == 'noun-unique' \
                else ((skipword(x) + random.choice(('', ' CLUB'))) \
                    if x.originclass == 'noun-mass' \
                    else (skipart(x) + ' CLUB')))), # Credit: mrob27
    Meme('THERE IS NO {}{} FOR THE {}',
        Rs['noun singular mass'],
        C((', SAID THE GLR,', 0.25), (', SAID RANDALL,', 0.33), ('', 1.0)),
        Rs['adjective']), # Credit: mrob27
    Meme('THERE\'S NO {} {} {}',
        C(('E**ISHNESS', 0.33), ('SADNESS', 0.5), ('CRYING', 1.0)),
        IndexMap(2, lambda x: x.wordforms[1]),
        Rs['place']), # Credit: mrob27
    Meme('THESE AREN\'T THE {} YOU\'RE {} FOR',
        Rs['noun plural'],
        Rs['verb cont']), # Credit: mrob27
    Meme('THEY\'RE TAKING THE {} TO {}!',
        R(('noun', n_plural), defaults=[('MOLPIES', 0.1)]),
        R('place', defaults=[('ISENGARD', 0.33), ('ᘝᓄᘈᖉᐣ', 0.33)])),
            # Credit: mrob27
    Meme('THE {} IS TOO CH*RPING {}',
        Rs['noun singular'],
        R('adjective', defaults=[('HIGH', 0.3)], postproc=skipart)),
            # Credit: mrob27
    Meme('THE {} PART OF {} IS {} IN YOUR {}',
        R(('adjective', a_super), defaults=[('MOLPIEST', 0.25)]),
        R(('verb-intr', v_cont), ('verb-tr', v_cont),
            defaults=[('MOLPYING UP', 0.2)]),
        Rs['noun plural mass'],
        Rs['noun singular mass']), # Credit: mrob27
    Meme('TO {0} OR NOT TO {0}, THAT IS THE QUESTION',
            Rs['verb inf']), # Credit: mrob27
    Meme('WE\'RE GONNA NEED A MORE {} {}',
        R('adjective', defaults=[('SEAISH', 0.5)], postproc=skipart),
        Rs['noun singular']), # Credit: mrob27
    Meme('WHAT HAPPENS IN {0}, STAYS IN {0}',
        Rs['place']),
    Meme('WITH {} LIKE THESE, WHO NEEDS {}?',
        Rs['noun plural'],
        Rs['noun plural mass']),
    Meme('WOW - SO {} - MANY {} - SUCH {} - VERY {} - MUCH {} - {}',
        Rs['noun singular'],
        Rs['adjective'],
        Rs['verb cont'],
        Rs['noun singular'],
        Rs['adjective'],
        C(('SEA IS BIG', 0.33), ('RIVER IS SMALL', 0.5),
            ('HAIL RANDALL', 1.0))), # Credit: mrob27
    Meme('XE WASN\'T THE {} {} {} {}',
        Rs['adjective superlative'],
        Rs['noun commodity'],
        IndexMap(3, lambda x: x.wordforms[1]),
        Rs['place']), # Credit: mrob27
    Meme('Y\'ALL GOT ANY MORE OF THEM {}?',
        Rs['noun plural']),
    Meme('YES {}, THERE _IS_ {}.',
        R('character', defaults=[('VIRGINIA', 0.2)], postproc=skipthe),
        R('noun-unique', 'noun-unique', 'noun')), # Credit: mrob27
    Meme('YOU CAN TAKE {0} OUT OF {1}, BUT YOU CAN\'T TAKE {1} OUT OF {0}',
        R('noun', 'noun-unique', 'character', postproc=lambda x: \
            ('THE ' + skipart(x)) if x.originclass.startswith('noun') \
            else x.stringform),
        R('place', selector=lambda x: x.wordforms[1] != 'ON')),
    Meme('YOU CAN\'T {} YOUR {} AND {} IT TOO',
        Rs['verb tr inf'],
        Rs['noun singular mass'],
        Rs['verb tr inf']), # Credit: mrob27
    Meme('_{} 2: ELECTRIC BAGALOO_',
        R(('noun', n_singular, n_plural), ('verb-tr', v_cont), 'noun-mass',
            postproc=skipart)), # Credit: mrob27
    Meme('_{0} HARD 2: {0} HARDER_',
        Rs['verb inf 1word']),
    Meme('{} AND {} AND {}, OH MY!',
        Rs['noun plural'],
        Rs['noun plural'],
        Rs['noun plural mass']), # Credit: mrob27
    Meme('{} AND {}, SITTING IN A {}TREE, {}',
        Rs['character'],
        Rs['character'],
        C(('NEAT ', 0.33), ('WOW', 0.5), ('', 1.0)),
        R(('verb-intr', v_cont), ('verb-tr', v_cont),
            selector=lambda x: x.stringform.endswith('ING'),
            postproc=lambda x: '-'.join(x.stringform))),
    Meme('{} FOR THE {} GOD! {} FOR THE {} THRONE!',
        Rs['noun plural mass'],
        IndexMap(0, lambda x: skipword(x.wordforms[n_singular]) \
                if x.originclass == 'noun' else x.stringform),
        Rs['noun plural mass'],
        IndexMap(2, lambda x: skipword(x.wordforms[n_singular]) \
                if x.originclass == 'noun' else x.stringform)),
    Meme('{} GRAB MY {}',
        Rs['character'],
        Rs['noun singular plural']),
    Meme('{}... IN SPACE!',
        Rs['noun plural mass character']),
    Meme('{} IS A {}',
        R(('verb-intr', v_cont), ('verb-tr', v_cont),
            selector=lambda x: x.stringform.endswith('ING')),
        R('noun', defaults=[('ART', 0.85)], selector=lambda x: \
            x.stringform.startswith('AN '), postproc=skipart)),
    Meme('{} IS A TIME LORD AND {} IS XIS TARDIS',
        Rs['character'],
        R('noun', 'noun-unique', 'place', postproc=lambda x: \
            ('THE ' + skipart(x)) if x.originclass.startswith('noun') \
            else x.stringform)),
    Meme('{} IS THE {} {} DESERVES, BUT NOT THE ONE IT NEEDS RIGHT NOW.',
        Rs['character'],
        R('noun', 'noun-unique', defaults=[('HERO', 0.3)], postproc=skipart),
        Rs['place']),
    Meme('{} IS {}. {}? NOT SO MUCH.',
        Rs['noun singular article mass character'],
        Rs['adjective'],
        Rs['noun singular article mass character']),
    Meme('{} IS "{} {}"... IF YOU KNOW WHAT I MEAN',
        Rs['character'],
        Rs['verb tr cont'],
        R('noun', 'noun-unique', 'place', 'character',
            postproc=lambda x: ('THE ' + skipart(x)) \
                if x.originclass == 'noun-unique' \
                else (('THE ' + skipart(x)) \
                    if x.originclass == 'noun' \
                    and random.choice((True, False)) else x.stringform))),
            # Credit: mrob27 and azule
    Meme('{} IT FROM {}. IT\'S THE ONLY WAY TO BE {}.',
        Rs['verb tr inf'],
        R('place', defaults=[('ORBIT', 0.85)]),
        R('adjective', defaults=[('SURE', 0.85)], postproc=skipart)),
    Meme('{} LIKE {}, {} LIKE {}',
        Rs['verb inf 1word'],
        R('noun', defaults=[('A FLUTTERBEE', 0.9)]),
        Rs['verb inf 1word'],
        R('noun', defaults=[('A BEESNAKE', 0.9)])),
    Meme('{0}MAN, {0}MAN, DOES WHATEVER {1} CAN',
        IndexMap(1, skipword),
        Rs['noun singular article']),
    Meme('{} RUN WHEN {} {} GOES TO {}',
        Rs['noun plural'],
        R('adjective', defaults=[('A GOOD', 0.25), ('A MOLPISH', 0.33)]),
        Rs['noun commodity'],
        R('place', defaults=[('WAR', 0.33)])),
    Meme('{0} THE {1}, {0} THE WORLD',
        R('verb-tr', defaults=[('SAVE', 0.67)]),
        Rs['noun singular plural mass']),
    Meme('{} USED {}! {}',
        Rs['character'],
        Rs['verb inf'],
        C(('IT\'S SUPER EFFECTIVE!', 0.5), ('IT\'S NOT VERY EFFECTIVE...',
            0.7), ('...BUT IT FAILED!', 1.0))),
    Meme('{}, WE HAVE {}SIGN THE LIKES OF WHICH EVEN {} HAS NEVER SEEN!',
        Rs['character'],
        R('noun', 'noun-unique', 'noun-mass',
            selector=lambda x: len(x.stringform.split()) == \
                    (2 if x.originclass == 'noun' else 1), postproc=skipart),
        C(('RANDALL', 0.33), ('GLR', 0.5), ('THE OTA', 1.0))),
    Meme('{} {} GONNA {}',
        R(('verb-intr', v_cont), ('verb-tr', v_cont),
            selector=lambda x: x.stringform.endswith('ING')),
        IndexMap(0, lambda x: erify(x.stringform) + 'S'),
        IndexMap(0, lambda x: x.wordforms[v_inf])), # Credit: mrob27
    Meme('{0}, {0}, MY KINGDOM FOR {0}',
            Rs['noun singular article']), # Credit: mrob27
    Meme('{} {} THE NEW {}',
        Rs['noun singular plural mass'],
        IndexMap(0, lambda x: 'ARE' if x.wordform == n_plural else 'IS'),
        IndexMap(0, lambda x: Rs['noun plural'].pick().stringform \
            if x.wordform == n_plural \
            else Rs['noun singular mass'].pick().stringform)),
            # Credit: mrob27
    Meme('{} - {} {}',
        C(('404', 0.33), ('408', 0.5), ('503', 1.0)),
        Rs['noun singular'],
        IndexMap(0, lambda x: {'404' : 'NOT FOUND', '408': 'REQUEST TIMEOUT',
            '503' : 'UNAVAILABLE'}[x.stringform]))
]

memes_meta_molpish = [
    Meme('MAYBE {} {}! MAYBE {} IS A METAPHOR FOR {}!',
        Rs['character'],
        C(('IS PREGNANT', 0.2), ('HAS CANCER', 0.25), ('HAD AN ABORTION',
            0.33), ('GOT MARRIED', 0.5), ('GOT DIVORCED', 1.0)),
        R('noun', 'noun-unique', 'noun-mass', defaults=[('THE OTC', 0.2)],
            postproc=lambda x: ('THE ' + skipart(x)) \
                if x.originclass != 'noun-mass' else skipart_mu(x)),
        R(('noun', 1), 'noun-mass', defaults=[('THIS', 0.1), ('THAT', 0.1)],
            postproc=skipart)),
    Meme('NI NI {} {} {}',
        C(('NI', 0.5), ('NI NI', 0.5), ('NI NI NI', 1.0)),
        R('noun', 'noun-unique', 'noun-mass', defaults=[('CHUPACABRA', 0.1)],
            postproc=skipart),
        R('noun', 'noun-unique', 'noun-mass',
            defaults=[('PING PONG BALL', 0.1)], postproc=skipart)),
    Meme('{0} {0} {0} {0} {0} {0} {0} {0} {0} {0} {0} {0} {1} {1}',
        R('noun', 'noun-unique', 'noun-mass', defaults=[('MOLPY', 0.1)],
            selector=lambda x: len(x.stringform.split()) == \
                (2 if x.originclass == 'noun' else 1), postproc=skipart),
        R('noun', 'noun-unique', 'noun-mass', defaults=[('GRAPEVINE', 0.1)],
            selector=lambda x: len(x.stringform.split()) == \
                (2 if x.originclass == 'noun' else 1), postproc=skipart))
                    # Credit: mrob27
]

memes_automeme_unmolpish = [
    Meme('BITCHES DON\'T KNOW ABOUT MY {}',
        Rs['noun singular plural mass']),
    Meme('HELLO. MY NAME IS {}. YOU KILLED MY {}. PREPARE TO {}.',
        Rs['character'],
        Rs['noun singular plural mass'],
        R('verb-intr', defaults=[('DIE', 0.9)])),
    Meme('I\'M {0}! I\'M {0}! SUCK MY DIIIIICK! I\'M {0}!',
        Rs['noun singular article character']),
    Meme('IT\'S OVER. {} IS FINISHED.',
        Rs['noun singular mass character']),
    Meme('{} HAET {}',
        R('character', defaults=[('LOLI', 0.5)]),
        Rs['noun singular plural mass character']),
    Meme('{} PIG DISGUSTING',
        R(('noun', n_singular, n_plural), 'noun-unique', 'character',
            postproc=lambda x: ('THE ' if x.originclass == 'noun-unique' \
                else '') + skipart_mu(x) + (' ARE' if x.wordform == 1 \
                else ' IS'))),
    Meme('{} IS THE {} THAT IS CH*RPING {}',
        R('noun', 'noun-unique', 'noun-mass', 'character',
            postproc=lambda x: (random.choice(('THIS ', 'THAT ')) \
                + skipword(x)) if x.originclass == 'noun' \
                else (('THE ' + skipword(x)) \
                    if x.originclass == 'noun-unique' else skipart_mu(x))),
        R('noun', 'noun-unique', 'character',
            defaults=[('BACONSEMENCOFFEEBABYCANCER', 0.8)], postproc=skipart),
        Rs['noun plural mass character']),
    Meme('{} KILLS {}',
        Rs['noun singular character'],
        Rs['noun plural character']),
    Meme('{0} YOUR CH*RPING {1}. {0} THEM.',
        Rs['verb tr inf'],
        Rs['noun plural'])
]

memes_other_unmolpish = [
    Meme('THAT\'S IT! I HAVE HAD IT WITH THESE MOLPYCHIRPING {} ON THIS '
            'MOLPYCHIRPING {}!',
        Rs['noun plural'],
        R('noun', 'noun-unique', defaults=[('RAFTCASTLE', 0.15)],
            postproc=skipart)),
    Meme('WELL CH*RP ME {} WITH {}',
        R('adverb'),
        Rs['noun singular article']),
    Meme('{} IS WHY WE CAN\'T HAVE {} THINGS.',
        R('noun', 'noun-unique', 'noun-mass', 'character',
            postproc=lambda x: (random.choice(('THIS ', 'THAT ')) \
                + skipword(x)) if x.originclass == 'noun' \
                else (('THE ' + skipword(x)) \
                if x.originclass == 'noun-unique' else skipart_mu(x))),
        C(('MOLPISH', 0.5), ('NICE', 1.0)))
]

def print_help(f=sys.stdout):
    print("""Usage: automome.py [option] [num]
[option] may be one of the following:
    -h, --help                      show this message and exit
    -V, --version                   show version and exit
    -c [cats], --categories=[cats]  generate memes from the categories [cats]
                                    (see below); the default is aom

[cats] may be any combination of the following letters:
    a       molpish memes from the original AUTOMEME (possibly OTTified)
    o       molpish other memes, i.e. non-AUTOMEME memes
    m       molpish meta-memes, i.e. memes originating in the OTT
    A       unmolpish AUTOMEME memes
    O       unmolpish other memes

[num] is the number of memes to generate; if it is missing or invalid, it is
assumed to be 1.""", file=f)

def print_version():
    print("""automome version 3.1, the OTTish AUTOMEME clone
Copyright (C) 2014 Penguin Development
Inspired by AUTOMEME (http://automeme.net/)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.""")

if __name__ == '__main__':
    n = 1
    memes = []
    opts, args = getopt.gnu_getopt(sys.argv[1:], 'hVc:', ('help', 'version',
        'categories='))

    for o, a in opts:
        if o in ('-h', '--help'):
            print_help()
            sys.exit(0)
        elif o in ('-V', '--version'):
            print_version()
            sys.exit(0)
        elif o in ('-c', '--categories'):
            if 'a' in a:
                memes += memes_automeme_molpish
            if 'o' in a:
                memes += memes_other_molpish
            if 'm' in a:
                memes += memes_meta_molpish
            if 'A' in a:
                memes += memes_automeme_unmolpish
            if 'O' in a:
                memes += memes_other_unmolpish

    if len(args) > 0:
        n = int(args[0])

    if n < 1:
        n = 1

    if len(memes) == 0:
        memes = memes_automeme_molpish + memes_other_molpish + \
                memes_meta_molpish

    for i in range(n):
        print(str(random.choice(memes)))
